
# coding: utf-8

# In[1]:


# This script reads in a spreadsheet containing dates & prices of two funds, calculates daily returns,
# calculates summary statistics of these returns, and writes the results to spreadsheets.
# The script is written in Python 3.6 and uses the following libraries: pandas (version 0.22.0), numpy (1.14.2),
# and datetime. You will also need to have installed the following libraries: scipy (1.0.0), xlrd (1.0.0), and
# xlsxwriter (0.9.6).


# In[2]:


# Import the necessary libraries
import pandas as pd
import numpy as np
import datetime


# In[3]:


# You can change the five variables below without having to adjust the rest of the script.

# Set the location & name of the spreadsheet containing the price data. Please note that Python uses the / symbol
# rather than \ to separate folders & sub-folders.
# The prices for both funds & the dates should be on the same sheet. The dates should be in column A; row 1 should
# contain the column title (for example, Date), and the dates should start on row 2. The prices should be in columns
# B &C; row 1 should contain the name of each fund, and the prices shold start in row 2. For every date, the should
# a price for both funds.
in_dir = 'C:/Analysis'
in_name = 'Equity-credit.xlsx'
# Set the location of the result spreadsheet, and how its name begins. The name of the results file will also contain
# the date.
out_dir = 'C:/Analysis'
out_base = 'Analysis of historical correlations_'
# Set the rolling periods (in days) whose correlations you want to calculate. You can amend, add, or delete periods.
# You should separate periods with a comma.
roll_days = (20, 60, 120)


# In[4]:


# The output file is an Excel spreadsheet with .xlsx as the extension (Excel 2007 or later). If you
# want to change to a different output format, you will need to make changes later in the script to
# ensure it works properly.
out_ext = '.xlsx'

# Make sure the directories end with /.
if in_dir[-1] == '/':
    in_file = in_dir + in_name
else:
    in_file = in_dir + '/' + in_name

if out_dir[-1] == '/':
    out_file = out_dir + out_base
else:
    out_file = out_dir + '/' + out_base

# Add the date (in YYYYMMDD format) to the end of the results file name.
out_file = out_file + datetime.datetime.strftime(datetime.datetime.today(), format = '%Y%m%d') + out_ext


# In[5]:


# Set the names of the sheets in the results file.
# If you want to change the order of the sheets, exclude particular sheets from the results file, or add
# new sheets, you will need to make changes later in the script to ensure it works properly.
out_sheets = ('Metrics',
              'Data Prices',
              'Data Returns and correlations')
# Set the names of the metrics that will be in the results file.
# If you want to change the order of the metrics, exclude particular metrics from the results file, or add
# new metrics, you will need to make changes later in the script to ensure it works properly.
var_names = ['Positive-positive days',
             'Positive-zero days',
             'Positive-negative days',
             'Zero-positive days',
             'Zero-zero days',
             'Zero-negative days',
             'Negative-positive days',
             'Negative-zero days',
             'Negative-negative days',
             'Positive-positive proportion',
             'Positive-zero proportion',
             'Positive-negative proportion',
             'Zero-positive proportion',
             'Zero-zero proportion',
             'Zero-negative proportion',
             'Negative-positive proportion',
             'Negative-zero proportion',
             'Negative-negative proportion',
             'Correlation of daily returns',
             'Correlation of the squares of daily returns']


# In[6]:


# Read in the input spreadsheet with the price data.
in_xl = pd.ExcelFile(in_file)


# In[7]:


# Create a dataframe with the relevant fund's prices. Ensure that these are the only three columns in the
# dataframe. Ensure that the column with dates is named Date and the columns with prices are named Price1
# and Price2
df_prices = pd.read_excel(in_file)
df_prices = df_prices.iloc[:, :3]
df_prices = df_prices.drop_duplicates()
original_columns = df_prices.columns
df_prices.columns = ['Date', 'Price1', 'Price2']


# In[8]:


# Sort the prices by date.
df_prices.sort_values(by = 'Date', ascending = True, inplace = True)
df_prices = df_prices.reset_index()
df_prices.drop('index', axis = 1, inplace = True)


# In[9]:


# Calculate the length of the period and number of pricing dates.
start_date = df_prices.Date.min()
end_date = df_prices.Date.max()
term_days = (end_date - start_date) / np.timedelta64(1, 'D')
term_years = term_days / 365.25
num_returns = len(df_prices) - 1
mean_days_per_year = num_returns / term_years


# In[10]:


# Create a dataframe of logarithmic returns from the dataframe of prices.
df_returns = df_prices.copy()
df_returns['Return1'] = np.log(df_returns.Price1) - np.log(df_returns.Price1.shift(1))
df_returns['Return2'] = np.log(df_returns.Price2) - np.log(df_returns.Price2.shift(1))
df_returns.drop(0, axis = 0, inplace = True)
df_returns.drop(['Price1', 'Price2'], axis = 1, inplace = True)


# In[11]:


# Calculate the return-sign matrix, that is, how often both funds had a positive return, how often
# one had a positive return while the other had a negative retrn, &c..
pos_pos_days = len(df_returns.loc[(df_returns.Return1 > 0) & (df_returns.Return2 > 0)])
pos_pos_prop = pos_pos_days / num_returns
pos_zero_days = len(df_returns.loc[(df_returns.Return1 > 0) & (df_returns.Return2 == 0)])
pos_zero_prop = pos_zero_days / num_returns
pos_neg_days = len(df_returns.loc[(df_returns.Return1 > 0) & (df_returns.Return2 < 0)])
pos_neg_prop = pos_neg_days / num_returns
zero_pos_days = len(df_returns.loc[(df_returns.Return1 == 0) & (df_returns.Return2 > 0)])
zero_pos_prop = zero_pos_days / num_returns
zero_zero_days = len(df_returns.loc[(df_returns.Return1 == 0) & (df_returns.Return2 == 0)])
zero_zero_prop = zero_zero_days / num_returns
zero_neg_days = len(df_returns.loc[(df_returns.Return1 == 0) & (df_returns.Return2 < 0)])
zero_neg_prop = zero_neg_days / num_returns
neg_pos_days = len(df_returns.loc[(df_returns.Return1 < 0) & (df_returns.Return2 > 0)])
neg_pos_prop = neg_pos_days / num_returns
neg_zero_days = len(df_returns.loc[(df_returns.Return1 < 0) & (df_returns.Return2 == 0)])
neg_zero_prop = neg_zero_days / num_returns
neg_neg_days = len(df_returns.loc[(df_returns.Return1 < 0) & (df_returns.Return2 < 0)])
neg_neg_prop = neg_neg_days / num_returns


# In[12]:


# Calculate the correlation of daily returns and the correlation of squared daily returns.
correl = df_returns['Return1'].corr(df_returns['Return2'])
df_returns['Squared return1'] = df_returns['Return1'] ** 2
df_returns['Squared return2'] = df_returns['Return2'] ** 2
squared_correl = df_returns['Squared return1'].corr(df_returns['Squared return2'])


# In[13]:


# Calculate the annualized mean & standard deviation of the returns for both funds. Use these
# to calculate the Sharpe ratios.
mean_return1 = df_returns.Return1.mean()
mean_return_annual1 = mean_return1 * mean_days_per_year
std_return1 = df_returns.Return1.std()
std_return_annual1 = std_return1 * mean_days_per_year ** 0.5
sharpe_ratio1 = mean_return_annual1 / std_return_annual1
mean_return2 = df_returns.Return2.mean()
mean_return_annual2 = mean_return2 * mean_days_per_year
std_return2 = df_returns.Return2.std()
std_return_annual2 = std_return2 * mean_days_per_year ** 0.5
sharpe_ratio2 = mean_return_annual2 / std_return_annual2


# In[14]:


# Loop through each chosen rolling period and calculate the rolling correlations between the
# returns and between the squared returns.
for roll_day in roll_days:
    col_name = 'Rolling ' + str(roll_day) + '-day correlation of returns'
    df_returns[col_name] = df_returns['Return1'].rolling(window = roll_day).corr(df_returns['Return2'])
    df_returns[col_name].fillna('', inplace = True)
    col_name = 'Rolling ' + str(roll_day) + '-day correlation of squared returns'
    df_returns[col_name] = df_returns['Squared return1'].rolling(window = roll_day).corr(df_returns['Squared return2'])
    df_returns[col_name].fillna('', inplace = True)


# In[15]:


# Include the Sharpe-ratio names in the list of metrics, and populate the metric values.
var_names = var_names + [original_columns[1] + ' Sharpe ratio', original_columns[2] + ' Sharpe ratio']
var_values = [pos_pos_days,
              pos_zero_days,
              pos_neg_days,
              zero_pos_days,
              zero_zero_days,
              zero_neg_days,
              neg_pos_days,
              neg_zero_days,
              neg_neg_days,
              pos_pos_prop,
              pos_zero_prop,
              pos_neg_prop,
              zero_pos_prop,
              zero_zero_prop,
              zero_neg_prop,
              neg_pos_prop,
              neg_zero_prop,
              neg_neg_prop,
              correl,
              squared_correl,
              sharpe_ratio1,
              sharpe_ratio2]


# In[16]:


# Tidy up the metrics dataframe.
df_metrics = pd.DataFrame(var_names).rename(columns = {0:'Metric'})
df_metrics['Value'] = pd.DataFrame(var_values)


# In[17]:


# Convert timestamps to dates.
df_prices.Date = df_prices.Date.dt.date
df_returns.Date = df_returns.Date.dt.date


# In[18]:


# Change the names of the columns for date & price back to those used in the input file.
df_prices.columns = original_columns
df_returns.rename(columns = {'Return1':original_columns[1] + ' return',
                             'Return2':original_columns[2] + ' return',
                             'Squared return1':original_columns[1] + ' squared return',
                             'Squared return2':original_columns[2] + ' squared return'}, inplace = True)


# In[19]:


# Write the results to a spreadsheet.
xl_writer = pd.ExcelWriter(out_file, engine = 'xlsxwriter')
dfs = [df_metrics, df_prices, df_returns]
df_num = 0


# In[20]:


# Loop through the types of result; set up a worksheet in the spreadsheet for each type, write the relevant
# results to the sheet, and set the width of each column in the worksheet so that it can accommodate the
# results and the name of the column.
for out_sheet_name in out_sheets:
    dfs[df_num].to_excel(xl_writer, index = False, sheet_name = out_sheet_name)
    out_sheet = xl_writer.sheets[out_sheet_name]
    col_num = 0
    
    for out_col in dfs[df_num].columns:
        column_len = dfs[df_num][[out_col]].astype(str).apply(lambda x: x.str.len()).max().max()
        header_len = len(dfs[df_num][[out_col]].astype(str).columns.values[0])
        out_sheet.set_column(col_num, col_num, column_len)
        
        if header_len > column_len:
            out_sheet.set_column(col_num, col_num, header_len)
        
        col_num = col_num + 1
    
    df_num = df_num + 1


# In[21]:


# Draw all the charts based on the data written to the spreadsheet.
# The script sets some parameters for each chart, but more parameters are available, so if you want to,
# for example, change the colour of a line or the font of text, you can add parameters & their values
# within the various sets of curly brackets below.
# Please see https://xlsxwriter.readthedocs.io/chart.html for more details.
workbook = xl_writer.book

# Draw a scatter plot of the daily returns of each fund. Fit & draw a line to these pairs of returns using
# ordinary least squares.
sheet_name = out_sheets[2]
worksheet = xl_writer.sheets[sheet_name]
    
chart = workbook.add_chart({'type': 'scatter'})
chart.add_series({'categories': [sheet_name, 1, 1, len(df_returns), 1],
                  'values': [sheet_name, 1, 2, len(df_returns), 2],
                  'trendline': {'type': 'linear','display_equation': True, 'display_r_squared': True}})
chart.set_x_axis({'name': original_columns[1],
                  'name_font': {'size': 14, 'bold': True},
                  'major_gridlines': {'visible': True, 'line': {'width': 0.25, 'dash_type': 'dash'}},
                  'major_tick_mark': 'none',
                  'minor_tick_mark': 'none',
                  'crossing': 'max'})
chart.set_y_axis({'name': original_columns[2],
                  'name_font': {'size': 14, 'bold': True},
                  'major_gridlines': {'visible': True, 'line': {'width': 0.25, 'dash_type': 'dash'}},
                  'major_tick_mark': 'none',
                  'minor_tick_mark': 'none',
                  'crossing': 'max'})
chart.set_title({'name': 'Correlation of daily returns'})
chart.set_legend({'position': 'none'})
chartsheet = workbook.add_chartsheet('Chart Correlation')
chartsheet.set_chart(chart)
    
chart = workbook.add_chart({'type': 'scatter'})
chart.add_series({'categories': [sheet_name, 1, 3, len(df_returns), 3],
                  'values': [sheet_name, 1, 4, len(df_returns), 4],
                  'trendline': {'type': 'linear','display_equation': True, 'display_r_squared': True}})
chart.set_x_axis({'name': original_columns[1],
                  'name_font': {'size': 14, 'bold': True},
                  'major_gridlines': {'visible': True, 'line': {'width': 0.25, 'dash_type': 'dash'}},
                  'major_tick_mark': 'none',
                  'minor_tick_mark': 'none',
                  'crossing': 'max'})
chart.set_y_axis({'name': original_columns[2],
                  'name_font': {'size': 14, 'bold': True},
                  'major_gridlines': {'visible': True, 'line': {'width': 0.25, 'dash_type': 'dash'}},
                  'major_tick_mark': 'none',
                  'minor_tick_mark': 'none',
                  'crossing': 'max'})
chart.set_title({'name': 'Correlation of squared daily returns'})
chart.set_legend({'position': 'none'})
chartsheet = workbook.add_chartsheet('Chart Squared correlation')
chartsheet.set_chart(chart)

# Loop through each chosen rolling period and draw a line chart of the rolling correlations between the
# returns and between the squared returns.
line_width = np.round((1500 / num_returns * 2.25) / 0.25, 0) * 0.25

if line_width > 3:
    line_width = 3
elif line_width < 0.25:
    line_width = 0.25

dummy_roll_day = 0

for roll_day in roll_days:
    chart = workbook.add_chart({'type': 'line'})
    chart.add_series({'categories': [sheet_name, roll_day, 0, len(df_returns), 0],
                      'values': [sheet_name, roll_day, 5 + dummy_roll_day * 2, len(df_returns), 5 + dummy_roll_day * 2],
                      'line': {'width': line_width}})
    chart.set_x_axis({'name': 'Date',
                      'name_font': {'size': 14, 'bold': True},
                      'major_gridlines': {'visible': True, 'line': {'width': 0.25, 'dash_type': 'dash'}},
                      'major_tick_mark': 'none',
                      'minor_tick_mark': 'none'})
    chart.set_y_axis({'name': 'Correlation',
                      'name_font': {'size': 14, 'bold': True},
                      'major_gridlines': {'visible': True, 'line': {'width': 0.25, 'dash_type': 'dash'}},
                      'major_tick_mark': 'none',
                      'minor_tick_mark': 'none'})
    chart.set_title({'name': 'Rolling ' + str(roll_day) + '-day correlation of daily returns'})
    chart.set_legend({'position': 'none'})
    chartsheet = workbook.add_chartsheet('Chart Rolling ' + str(roll_day) + '-day corr')
    chartsheet.set_chart(chart)
    
    chart = workbook.add_chart({'type': 'line'})
    chart.add_series({'categories': [sheet_name, roll_day, 0, len(df_returns), 0],
                      'values': [sheet_name, roll_day, 6 + dummy_roll_day * 2, len(df_returns), 6 + dummy_roll_day * 2],
                      'line': {'width': line_width}})
    chart.set_x_axis({'name': 'Date',
                      'name_font': {'size': 14, 'bold': True},
                      'major_gridlines': {'visible': True, 'line': {'width': 0.25, 'dash_type': 'dash'}},
                      'major_tick_mark': 'none',
                      'minor_tick_mark': 'none'})
    chart.set_y_axis({'name': 'Correlation',
                      'name_font': {'size': 14, 'bold': True},
                      'major_gridlines': {'visible': True, 'line': {'width': 0.25, 'dash_type': 'dash'}},
                      'major_tick_mark': 'none',
                      'minor_tick_mark': 'none'})
    chart.set_title({'name': 'Rolling ' + str(roll_day) + '-day correlation of squared daily returns'})
    chart.set_legend({'position': 'none'})
    chartsheet = workbook.add_chartsheet('Chart Rolling ' + str(roll_day) + '-day sqrd corr')
    chartsheet.set_chart(chart)
    
    dummy_roll_day = dummy_roll_day + 1


# In[22]:


# Save the spreadsheet.
xl_writer.save()

