
# coding: utf-8

# In[1]:


# This script reads in a spreadsheet containing dates & prices of various funds, calculates daily returns,
# calculates summary statistics of these returns, and writes the results to spreadsheets.
# The script is written in Python 3.6 and uses the following libraries: pandas (version 0.22.0), numpy (1.14.2)
# statsmodel (0.8.0), and datetime. You will also need to have installed the following libraries: scipy (1.0.0),
# xlrd (1.0.0), and xlsxwriter (0.9.6).


# In[2]:


# Import the necessary libraries
import pandas as pd
import numpy as np
import statsmodels.formula.api as sm
import datetime


# In[3]:


# You can change the ten variables below without having to adjust the rest of the script.

# Set the location & name of the spreadsheet containing the price data. Please note that Python uses the / symbol
# rather than \ to separate folders & sub-folders.
# Each fund's prices should be on a separate sheet, and the sheet's name should reflect the name of the fund.
# For each fund, column A should contain the dates, and column B should contain the prices.
# Row 1 should contain the column titles (for example, Date & Price), and the dates & prices should start on row 2.
in_dir = 'C:/Analysis'
in_name = 'Prices.xlsx'
# Set the location of the results spreadsheets, and how their names begin.
# The names of the results files will also contain the name of the sheets that their prices are on.
out_dir = 'C:/Analysis'
out_base = 'Analysis of historical returns_'
# Set the rolling periods (in days) whose returns you want to calculate. You can amend, add, or delete periods.
# You should separate periods with a comma.
roll_days = (5, 20)
# Set the proportions of the total return for which you want to calculate the minimum number of days that
# account for it. You can amend, add, or delete proportions. You should separate proportions with a comma.
total_percs = (90, 80, 75)
# Set the rolling periods (in years) for which you want to determine the proportion with positive returns. And set 
# whether you want to show the rolling annualized returns for these periods (True or False). You can amend, add,
# or delete periods. You should separate periods with a comma.
roll_years = (1, 3, 5, 7)
show_ann_roll_years = True
# Set the number of bins when determining the distribution of daily returns, that is, how many slices of
# equal width the range of daily returns will be split into.
num_bins = 35
# There will be a plot of Sharpe ratios since the start date. The early Sharpe ratios will jump around a bit because
# of the small number of data points. So that these early ratios do not distort the plot too much, you can choose
# how many of the earliest ratios to exclude.
sharpe_exclude = 250


# In[4]:


# The output files are Excel spreadsheets with .xlsx as the extension (Excel 2007 or later). If you
# want to change to a different output format, you will need to make changes later in the script to
# ensure it works properly.
out_ext = '.xlsx'

# Make sure the directories end with /.
if in_dir[-1] == '/':
    in_file = in_dir + in_name
else:
    in_file = in_dir + '/' + in_name

if out_dir[-1] == '/':
    out_file = out_dir + out_base
else:
    out_file = out_dir + '/' + out_base

# Add the date (in YYYYMMDD format) to the end of the results file name.
out_ext = '_' + datetime.datetime.strftime(datetime.datetime.today(), format = '%Y%m%d') + out_ext


# In[5]:


# Set the names of the sheets in each results file.
# If you want to change the order of the sheets, exclude particular sheets from the results file, or add
# new sheets, you will need to make changes later in the script to ensure it works properly.
out_sheets = ('Metrics',
              'Data Prices',
              'Data Returns',
              'Data Paired returns',
              'Data Ordered returns',
              'Data Binned returns',
              'Data Sharpe ratios')
# Set the names of the metrics that will be in each results file.
# If you want to change the order of the metrics, exclude particular metrics from the results file, or add
# new metrics, you will need to make changes later in the script to ensure it works properly.
var_names = ['Term (years)',
             'Maximum gap between prices (days)',
             'Minimum gap between prices (days)',
             'Mean gap between prices (days)',
             'Standard deviation gap between prices (days)',
             'Mean daily return',
             'Mean daily return (annualized)',
             'Median daily return',
             'Median daily return (annualized)',
             'Standard deviation daily return',
             'Standard deviation daily return (annualized)',
             'Skewness daily return',
             'Kurtosis daily return',
             'Maximum standardized daily return (SDs above the mean)',
             'Minimum standardized daily return (SDs below the mean)',
             'Maximum 5-day return',
             'Maximum 10-day return',
             'Slope of paired returns',
             'Number of days with positive return',
             'Proportion of days with positive return',
             'Mean positive return',
             'Number of days with negative return',
             'Proportion of days with negative return',
             'Mean negative return',
             'Number of days with zero return',
             'Proportion of days with zero return',
             'Omega ratio',
             'Maximum peak-to-trough fall',
             'Maximum peak-to-trough fall (non-logarithmic)']


# In[6]:


# Add the names of the proportion-of-total-return and rolling-positive-return-proportion metrics.
for total_perc in total_percs:
    var_names.append('Number of days accounting for ' + str(total_perc) + '% of total return')

for total_perc in total_percs:
    var_names.append('Proportion of days accounting for ' + str(total_perc) + '% of total return')

for total_perc in total_percs:
    var_names.append('Proportion of positive-return days accounting for ' + str(total_perc) + '% of positive return')

for total_perc in total_percs:
    var_names.append('Proportion of negative-return days accounting for ' + str(total_perc) + '% of negative return')

for roll_year in roll_years:
    var_names.append('Proportion of rolling ' + str(roll_year) + '-year returns positive')


# In[7]:


# Read in the input spreadsheet with the price data.
in_xl = pd.ExcelFile(in_file)

# Loop through each sheet in the spreadsheet.
for in_sheet_name in in_xl.sheet_names:
# Create a dataframe with the relevant fund's prices. Ensure that these are the only two columns in the
# dataframe. Ensure that the column with dates is named Date and the column with prices is named Price. 
    df_prices = pd.read_excel(in_xl, sheet_name = in_sheet_name)
    df_prices = df_prices.iloc[:, :2]
    df_prices = df_prices.drop_duplicates()
    original_columns = df_prices.columns
    df_prices.columns = ['Date', 'Price']

# Sort the prices by date.
    df_prices.sort_values(by = 'Date', ascending = True, inplace = True)
    df_prices = df_prices.reset_index()
    df_prices.drop('index', axis = 1, inplace = True)

# Calculate the length of the period and number of pricing dates.
    start_date = df_prices.Date.min()
    end_date = df_prices.Date.max()
    term_days = (end_date - start_date) / np.timedelta64(1, 'D')
    term_years = term_days / 365.25
    num_returns = len(df_prices) - 1
    mean_days_per_year = num_returns / term_years
    
# Calculate summary stats for gaps between pricing dates.    
    max_price_gap_days = ((df_prices.Date - df_prices.Date.shift(1)) / np.timedelta64(1, 'D')).max()
    min_price_gap_days = ((df_prices.Date - df_prices.Date.shift(1)) / np.timedelta64(1, 'D')).min()
    mean_price_gap_days = ((df_prices.Date - df_prices.Date.shift(1)) / np.timedelta64(1, 'D')).mean()
    std_price_gap_days = ((df_prices.Date - df_prices.Date.shift(1)) / np.timedelta64(1, 'D')).std()
    
# Create a dataframe of logarithmic returns from the dataframe of prices.
    df_returns = df_prices.copy()
    df_returns['Return'] = np.log(df_returns.Price) - np.log(df_returns.Price.shift(1))
    df_returns.drop(0, axis = 0, inplace = True)
    df_returns.drop('Price', axis = 1, inplace = True)
    
# Calculate summary stats for the daily returns.
    mean_return = df_returns.Return.mean()
    mean_return_annual = mean_return * mean_days_per_year
    median_return = df_returns.Return.median()
    median_return_annual = median_return * mean_days_per_year
    std_return = df_returns.Return.std()
    std_return_annual = std_return * mean_days_per_year ** 0.5
    skew_return = df_returns.Return.skew()
    kurt_return = df_returns.Return.kurtosis()
    max_return_standard = (df_returns.Return.max() - mean_return) / std_return
    min_return_standard = (df_returns.Return.min() - mean_return) / std_return
    max_return_5 = df_returns.Return.rolling(5).sum().max()
    max_return_10 = df_returns.Return.rolling(10).sum().max()
    
# Populate the prices of any non-pricing dates by using the most recent prices.
    df_dates = pd.DataFrame(
        pd.date_range(start = df_prices.Date.min(), end = df_prices.Date.max(), freq = '1D')).rename(columns = {0:'Date'})
    df_all_prices = df_dates.merge(df_prices, on = 'Date', how = 'left')
    df_all_prices.Price.fillna(method = 'ffill', inplace = True)
    df_all_prices.set_index('Date', inplace = True)

# Calculate the minimum number of days whose returns make up the chosen proportions of the total return.
# Express this as a number of days & as a proportion of the total number of days.
    df_pos_returns = df_returns.loc[df_returns.Return > 0]
    num_pos_returns = len(df_pos_returns)
    df_neg_returns = df_returns.loc[df_returns.Return < 0]
    num_neg_returns = len(df_neg_returns)
    total_return = df_returns.Return.sum()
    total_pos_return = df_pos_returns.Return.sum()
    total_neg_return = df_neg_returns.Return.sum()
    df_returns.sort_values(by = 'Return', ascending = False, inplace = True)
    df_pos_returns.sort_values(by = 'Return', ascending = False, inplace = True)
    df_neg_returns.sort_values(by = 'Return', ascending = True, inplace = True)
    
    total_perc_num_days = []
    total_perc_num_days_prop = []
    pos_perc_num_days_prop = []
    neg_perc_num_days_prop = []
    
    for total_perc in total_percs:
        num_days = len(df_returns.loc[df_returns.Return.cumsum() < total_return * total_perc / 100]) + 1
        num_days_prop = num_days / num_returns
        total_perc_num_days.append(num_days)
        total_perc_num_days_prop.append(num_days_prop)
        num_pos_days = len(df_pos_returns.loc[df_pos_returns.Return.cumsum() < total_pos_return * total_perc / 100]) + 1
        num_pos_days_prop = num_pos_days / num_pos_returns
        pos_perc_num_days_prop.append(num_pos_days_prop)
        num_neg_days = len(df_neg_returns.loc[df_neg_returns.Return.cumsum() > total_neg_return * total_perc / 100]) + 1
        num_neg_days_prop = num_neg_days / num_neg_returns
        neg_perc_num_days_prop.append(num_neg_days_prop)
    
# Create a dataframe containing just the daily returns. This has already been ordered from highest to lowest.
# Add a field equal to the daily returns ordered the opposite way, that is, from lowest to highest. The
# dataframe now contains pairs of returns: highest & lowest, second-highest & second-lowest, third-highest
# & third-lowest, and so on. Include only pairs where one return is positive and one is negative.
# Fit a line to these pairs of returns using ordinary least squares, and calculate the slope of this line.
    df_paired_returns = df_returns[['Return']].copy().reset_index()
    df_paired_returns['Paired'] = df_returns.sort_values(by = 'Return', ascending = True).reset_index()['Return']
    df_paired_returns.drop('index', axis = 1, inplace = True)
    df_paired_returns = df_paired_returns.loc[(df_paired_returns['Return'] > 0) & (df_paired_returns['Paired'] < 0)]
    ols = sm.ols(formula = 'Return ~ Paired', data = df_paired_returns).fit()
    paired_return_slope = ols.params[1]
    
# Calculate the number of days with positive/negative returns and the mean return on both types of day.
# Calculate the Omega ratio, that is, the total return on positive days divided by the total return on
# negative days.
    num_pos_days = num_pos_returns
    num_pos_days_prop = num_pos_days / num_returns
    mean_pos_days = df_returns.loc[df_returns.Return > 0, 'Return'].mean()
    num_neg_days = num_neg_returns
    num_neg_days_prop = num_neg_days / num_returns
    mean_neg_days = df_returns.loc[df_returns.Return < 0, 'Return'].mean()
    num_zero_days = len(df_returns.loc[df_returns.Return == 0])
    num_zero_days_prop = num_zero_days / num_returns
    omega_ratio = -df_returns.loc[df_returns.Return > 0, 'Return'].sum() / df_returns.loc[df_returns.Return < 0, 'Return'].sum()
    
    df_returns.sort_values(by = 'Date', ascending = True, inplace = True)
    
# Loop through each chosen rolling period (in years) and calculate the proportion of each length of
# period with positive returns.
# If you have chosen to show the rolling annualized returns for each chosen period, then calculate
# these returns.
    roll_props = []
    
    for roll_year in roll_years:
        if roll_year <= term_years:
            first_roll_date = df_prices.Date.min()
            last_date = df_prices.Date.max()
            
            if show_ann_roll_years:
                roll_year_col = 'Rolling annualized ' + str(roll_year) + '-year return'
                df_returns[roll_year_col] = np.nan
            else:
                num_pos_periods = 0
                num_non_pos_periods = 0
            
# Find the date exactly roll_year years before the last pricing date. If the last pricing date is
# February 29th and roll_year is not divisible by 4, then there will be no date exactly roll_year
# years before the last pricing date, so use February 28th instead.
            try:
                last_roll_date = datetime.datetime(last_date.year - roll_year,
                                                   last_date.month,
                                                   last_date.day)
            except:
                last_roll_date = datetime.datetime(last_date.year - roll_year,
                                                   last_date.month,
                                                   last_date.day - 1)
            
# Loop through every date (whether or not it was a pricing date) and compare the price applicable
# to that date with the price applicable exactly roll_year years later to determine whether the
# return over that particular period was positive.
            for start_roll_date in pd.date_range(start = first_roll_date, end = last_roll_date, freq = '1D'):
                start_price = df_all_prices.Price.loc[start_roll_date]
                
# Find the date exactly roll_year years after the start date. If the start date is February
# 29th and roll_year is not divisible by 4, then there will be no date exactly roll_year years
# after the start date, so use February 28th instead.
                try:
                    end_roll_date = datetime.datetime(start_roll_date.year + roll_year,
                                                      start_roll_date.month,
                                                      start_roll_date.day)
                except:
                    end_roll_date = datetime.datetime(start_roll_date.year + roll_year,
                                                      start_roll_date.month,
                                                      start_roll_date.day - 1)
                
                end_price = df_all_prices.Price.loc[end_roll_date]
                
                if show_ann_roll_years:
                    df_returns.loc[
                        df_returns.Date == end_roll_date, roll_year_col] = (np.log(end_price) - np.log(start_price)) / roll_year
                else:
                    if end_price > start_price:
                        num_pos_periods = num_pos_periods + 1
                    else:
                        num_non_pos_periods = num_non_pos_periods + 1
                
            if show_ann_roll_years:
                num_pos_periods = len(df_returns.loc[df_returns[roll_year_col] > 0])
                num_non_pos_periods = len(df_returns.loc[df_returns[roll_year_col] <= 0])
            
            roll_prop = num_pos_periods / (num_pos_periods + num_non_pos_periods)
        else:
            roll_prop = 'n/a'
        
        roll_props.append(roll_prop)
    
# For each pricing date, if the price on that date is the highest so far, then find the lowest price that
# follows it and calculate the peak-to-trough return. The lowest such return after looping through all dates
# is the maximum peak-to-trough fall.
    old_peak = 0
    max_peak_to_trough = 0
    
    for dummy_date in np.arange(len(df_prices) - 1):
        peak = df_prices.Price[dummy_date]
        
        if peak > old_peak:
            old_peak = peak
            trough = df_prices.Price[dummy_date + 1:].min()
            peak_to_trough = np.log(peak) - np.log(trough)
            if max_peak_to_trough < peak_to_trough: max_peak_to_trough = peak_to_trough
    
    if max_peak_to_trough < 0: max_peak_to_trough = 0
    
# Calculate the non-logarithmic equivalent of the maximum peak-to-trough fall.
    max_peak_to_trough_non_log = 1 - np.exp(-max_peak_to_trough)

# Calculate the annualized returns & standard deviations from the start date.
# Calculate the Sharpe ratio from these.
    df_sharpe = df_returns[['Date', 'Return']]
    df_sharpe['Days since start'] = (df_sharpe.Date - start_date) / np.timedelta64(1, 'D')
    df_sharpe['Annualized return since start'] = df_sharpe.Return.cumsum() * 365.25 / df_sharpe['Days since start']
    df_sharpe['Annualized standard deviation since start'] = df_sharpe.Return.rolling(
        len(df_sharpe), min_periods = 2).std() * mean_days_per_year ** 0.5
    df_sharpe['Sharpe ratio since start'] = df_sharpe['Annualized return since start'] / df_sharpe[
        'Annualized standard deviation since start']
    df_sharpe['Annualized standard deviation since start'].fillna('', inplace = True)
    df_sharpe['Sharpe ratio since start'].fillna('', inplace = True)
    df_sharpe.drop(['Return', 'Days since start'], axis = 1, inplace = True)

# Populate the metric values.
    var_values = [term_years,
                  max_price_gap_days,
                  min_price_gap_days,
                  mean_price_gap_days,
                  std_price_gap_days,
                  mean_return,
                  mean_return_annual,
                  median_return,
                  median_return_annual,
                  std_return,
                  std_return_annual,
                  skew_return,
                  kurt_return,
                  max_return_standard,
                  min_return_standard,
                  max_return_5,
                  max_return_10,
                  paired_return_slope,
                  num_pos_days,
                  num_pos_days_prop,
                  mean_pos_days,
                  num_neg_days,
                  num_neg_days_prop,
                  mean_neg_days,
                  num_zero_days,
                  num_zero_days_prop,
                  omega_ratio,
                  max_peak_to_trough,
                  max_peak_to_trough_non_log]
    
    for num_day in total_perc_num_days:
        var_values.append(num_day)
    
    for num_day_prop in total_perc_num_days_prop:
        var_values.append(num_day_prop)
        
    for num_day_prop in pos_perc_num_days_prop:
        var_values.append(num_day_prop)
        
    for num_day_prop in neg_perc_num_days_prop:
        var_values.append(num_day_prop)
        
    for roll_prop in roll_props:
        var_values.append(roll_prop)
    
# Tidy up the metrics dataframe.
    df_metrics = pd.DataFrame(var_names).rename(columns = {0:'Metric'})
    df_metrics['Value'] = pd.DataFrame(var_values)
    
# Convert timestamps to dates.
    df_returns.Date = df_returns.Date.dt.date
    df_prices.Date = df_prices.Date.dt.date
    df_sharpe.Date = df_sharpe.Date.dt.date
    
# Create a dataframe containing just the daily returns, ordered from lowest to highest.
    df_ordered_returns = df_returns.drop('Date', axis = 1).sort_values(by = 'Return', ascending = True).copy()
    
# Calculate the returns, annualized returns, & annualized standard deviations for each chosen
# rolling period (in days)
    for roll_day in roll_days:
        roll_sum = df_returns.Return.rolling(roll_day).sum()
        roll_day_col = 'Rolling ' + str(roll_day) + '-day return'
        df_returns[roll_day_col] = roll_sum
        roll_day_col = 'Rolling ' + str(roll_day) + '-day annual return'
        df_returns[roll_day_col] = roll_sum * mean_days_per_year / roll_day
        roll_day_col = 'Rolling ' + str(roll_day) + '-day annual standard deviation'
        df_returns[roll_day_col] = df_returns.Return.rolling(roll_day).std() * mean_days_per_year ** 0.5
    
# Calculate the PDF of daily returns by counting how many returns are in each bin.
# Calculate the CDF of daily returns using the cumulative sum of how many returns are in each bin and
# expressing this as a proportion of the total number of returns.
    df_binned_returns = pd.DataFrame(
        pd.cut(
            df_ordered_returns['Return'], num_bins)).groupby('Return').size().reset_index().rename(
        columns = {'Return':'Mid return in bin (bps)', 0:'Frequency'})
    df_binned_returns['Mid return in bin (bps)'] = df_binned_returns['Mid return in bin (bps)'].cat.categories.mid
    df_binned_returns['Mid return in bin (bps)'] = df_binned_returns['Mid return in bin (bps)'] * 100
    df_binned_returns['Cumulative proportion'] = df_binned_returns.Frequency.cumsum() / num_returns    
    
# Change the names of the columns for date & price back to those used in the input file.
    df_prices.columns = original_columns
    
# Write the results to a spreadsheet.
    xl_writer = pd.ExcelWriter(out_file + in_sheet_name + out_ext, engine = 'xlsxwriter')
    dfs = [df_metrics, df_prices, df_returns, df_paired_returns, df_ordered_returns, df_binned_returns, df_sharpe]
    df_num = 0
    
# Loop through the types of result; set up a worksheet in the spreadsheet for each type, write the relevant
# results to the sheet, and set the width of each column in the worksheet so that it can accommodate the
# results and the name of the column.
    for out_sheet_name in out_sheets:
        dfs[df_num].to_excel(xl_writer, index = False, sheet_name = out_sheet_name)
        out_sheet = xl_writer.sheets[out_sheet_name]
        col_num = 0
        
        for out_col in dfs[df_num].columns:
            column_len = dfs[df_num][[out_col]].astype(str).apply(lambda x: x.str.len()).max().max()
            header_len = len(dfs[df_num][[out_col]].astype(str).columns.values[0])
            out_sheet.set_column(col_num, col_num, column_len)
            
            if header_len > column_len:
                out_sheet.set_column(col_num, col_num, header_len)
            
            col_num = col_num + 1
        
        df_num = df_num + 1
    
# Draw all the charts based on the data written to each spreadsheet.
# The script sets some parameters for each chart, but more parameters are available, so if you want to,
# for example, change the colour of a line or the font of text, you can add parameters & their values
# within the various sets of curly brackets below.
# Please see https://xlsxwriter.readthedocs.io/chart.html for more details.
    workbook = xl_writer.book
    
# Draw a line chart of how prices have evolved over time.
    sheet_name = out_sheets[1]
    worksheet = xl_writer.sheets[sheet_name]
    
    min_price = df_prices.iloc[:, 1].min()
    max_price = df_prices.iloc[:, 1].max()
# Pick a sensible minimum & maximum price to plot, based on the size of the prices; for example,
# if the lowest price in the data is 9.3, start the chart at 9, but if the lowest price is 930,
# start the chart at 900.
    order_of_magnitude = 10 ** np.floor(np.log(min_price) / np.log(10))
    chart_min_price = int(min_price / order_of_magnitude) * order_of_magnitude
    chart_max_price = int(max_price / order_of_magnitude + 1) * order_of_magnitude
    chart = workbook.add_chart({'type': 'line'})
    chart.add_series({'categories': [sheet_name, 1, 0, len(df_prices), 0],
                      'values': [sheet_name, 1, 1, len(df_prices), 1]})
    chart.set_x_axis({'name': original_columns[0],
                      'name_font': {'size': 14, 'bold': True},
                      'major_gridlines': {'visible': True, 'line': {'width': 0.25, 'dash_type': 'dash'}},
                      'major_tick_mark': 'none',
                      'minor_tick_mark': 'none'})
    chart.set_y_axis({'name': original_columns[1],
                      'name_font': {'size': 14, 'bold': True},
                      'major_gridlines': {'visible': True, 'line': {'width': 0.25, 'dash_type': 'dash'}},
                      'major_tick_mark': 'none',
                      'minor_tick_mark': 'none',
                      'min': chart_min_price,
                      'max': chart_max_price})
    chart.set_title({'name': 'Prices over time'})
    chart.set_legend({'position': 'none'})
    chartsheet = workbook.add_chartsheet('Chart Prices')
    chartsheet.set_chart(chart)

# Draw a line chart of daily returns over time.
    sheet_name = out_sheets[2]
    worksheet = xl_writer.sheets[sheet_name]
    line_width = np.round((1500 / num_returns * 2.25) / 0.25, 0) * 0.25
    
    if line_width > 3:
        line_width = 3
    elif line_width < 0.25:
        line_width = 0.25
    
    chart = workbook.add_chart({'type': 'line'})
    chart.add_series({'categories': [sheet_name, 1, 0, len(df_returns), 0],
                      'values': [sheet_name, 1, 1, len(df_returns), 1],
                      'line': {'width': line_width}})
    chart.set_x_axis({'name': 'Date',
                      'name_font': {'size': 14, 'bold': True},
                      'major_gridlines': {'visible': True, 'line': {'width': 0.25, 'dash_type': 'dash'}},
                      'major_tick_mark': 'none',
                      'minor_tick_mark': 'none'})
    chart.set_y_axis({'name': 'Return',
                      'name_font': {'size': 14, 'bold': True},
                      'major_gridlines': {'visible': True, 'line': {'width': 0.25, 'dash_type': 'dash'}},
                      'major_tick_mark': 'none',
                      'minor_tick_mark': 'none'})
    chart.set_title({'name': 'Daily returns over time'})
    chart.set_legend({'position': 'none'})
    chartsheet = workbook.add_chartsheet('Chart Returns')
    chartsheet.set_chart(chart)
    
# If you have chosen to show rolling annualized returns, then draw line charts for each rolling period (in years).
    if show_ann_roll_years:
        dummy_roll_year = 0
        
        for roll_year in roll_years:
            first_roll_row = df_returns.loc[df_returns.Date == df_returns.loc[
                ~df_returns['Rolling annualized ' + str(roll_year) + '-year return'].apply(np.isnan), 'Date'].min()].index[0] + 1
            chart = workbook.add_chart({'type': 'line'})
            chart.add_series({'categories': [sheet_name, first_roll_row, 0, len(df_returns), 0],
                              'values': [sheet_name, first_roll_row, 2 + dummy_roll_year, len(df_returns), 2 + dummy_roll_year],
                              'line': {'width': line_width}})
            chart.set_x_axis({'name': 'Date',
                              'name_font': {'size': 14, 'bold': True},
                              'major_gridlines': {'visible': True, 'line': {'width': 0.25, 'dash_type': 'dash'}},
                              'major_tick_mark': 'none',
                              'minor_tick_mark': 'none'})
            chart.set_y_axis({'name': 'Return',
                              'name_font': {'size': 14, 'bold': True},
                              'major_gridlines': {'visible': True, 'line': {'width': 0.25, 'dash_type': 'dash'}},
                              'major_tick_mark': 'none',
                              'minor_tick_mark': 'none'})
            chart.set_title({'name': 'Rolling annualized ' + str(roll_year) + '-year returns'})
            chart.set_legend({'position': 'none'})
            chartsheet = workbook.add_chartsheet('Chart Rolling ' + str(roll_year) + '-year returns')
            chartsheet.set_chart(chart)
            
            dummy_roll_year = dummy_roll_year + 1
    
# For each rolling period (in days), draw line charts of rolling returns & annualized standard
# deviations over time.
    dummy_roll_day = 0
    first_value_column = 2
    if show_ann_roll_years: first_value_column = first_value_column + len(roll_years)
    
    for roll_day in roll_days:
        chart = workbook.add_chart({'type': 'line'})
        chart.add_series({'categories': [sheet_name, 1, 0, len(df_returns), 0],
                          'values': [sheet_name, 1, first_value_column + dummy_roll_day * 3,
                                     len(df_returns), first_value_column + dummy_roll_day * 3],
                          'line': {'width': line_width}})
        chart.set_x_axis({'name': 'Date',
                          'name_font': {'size': 14, 'bold': True},
                          'major_gridlines': {'visible': True, 'line': {'width': 0.25, 'dash_type': 'dash'}},
                          'major_tick_mark': 'none',
                          'minor_tick_mark': 'none'})
        chart.set_y_axis({'name': 'Return',
                          'name_font': {'size': 14, 'bold': True},
                          'major_gridlines': {'visible': True, 'line': {'width': 0.25, 'dash_type': 'dash'}},
                          'major_tick_mark': 'none',
                          'minor_tick_mark': 'none'})
        chart.set_title({'name': 'Rolling ' + str(roll_day) + '-day returns'})
        chart.set_legend({'position': 'none'})
        chartsheet = workbook.add_chartsheet('Chart Rolling ' + str(roll_day) + '-day returns')
        chartsheet.set_chart(chart)

        chart = workbook.add_chart({'type': 'line'})
        chart.add_series({'categories': [sheet_name, 1, 0, len(df_returns), 0],
                          'values': [sheet_name, 1, first_value_column + 2 + dummy_roll_day * 3,
                                     len(df_returns), first_value_column + 2 + dummy_roll_day * 3],
                          'line': {'width': line_width}})
        chart.set_x_axis({'name': 'Date',
                          'name_font': {'size': 14, 'bold': True},
                          'major_gridlines': {'visible': True, 'line': {'width': 0.25, 'dash_type': 'dash'}},
                          'major_tick_mark': 'none',
                          'minor_tick_mark': 'none'})
        chart.set_y_axis({'name': 'Return',
                          'name_font': {'size': 14, 'bold': True},
                          'major_gridlines': {'visible': True, 'line': {'width': 0.25, 'dash_type': 'dash'}},
                          'major_tick_mark': 'none',
                          'minor_tick_mark': 'none'})
        chart.set_title({'name': 'Rolling ' + str(roll_day) + '-day annualized standard deviations'})
        chart.set_legend({'position': 'none'})
        chartsheet = workbook.add_chartsheet('Chart Rolling ' + str(roll_day) + '-day std dev')
        chartsheet.set_chart(chart)
        
        dummy_roll_day = dummy_roll_day + 1
    
# Draw a scatter plot of the highest daily return against the lowest, the second-highest against the
# second-lowest, and so on. Fit & draw a line to these pairs of returns using ordinary least squares.
    sheet_name = out_sheets[3]
    worksheet = xl_writer.sheets[sheet_name]
    
    chart = workbook.add_chart({'type': 'scatter'})
    chart.add_series({'categories': [sheet_name, 1, 1, len(df_paired_returns), 1],
                      'values': [sheet_name, 1, 0, len(df_paired_returns), 0],
                      'trendline': {'type': 'linear',
                                    'display_equation': True}})
    chart.set_x_axis({'major_gridlines': {'visible': True, 'line': {'width': 0.25, 'dash_type': 'dash'}},
                      'major_tick_mark': 'none',
                      'minor_tick_mark': 'none',
                      'crossing': 'max'})
    chart.set_y_axis({'major_gridlines': {'visible': True, 'line': {'width': 0.25, 'dash_type': 'dash'}},
                      'major_tick_mark': 'none',
                      'minor_tick_mark': 'none'})
    chart.set_title({'name': 'High returns vs low returns'})
    chart.set_legend({'position': 'none'})
    chartsheet = workbook.add_chartsheet('Chart High vs low returns')
    chartsheet.set_chart(chart)
    
# Plot an area chart of returns ranked from lowest to highest.
    sheet_name = out_sheets[4]
    worksheet = xl_writer.sheets[sheet_name]
    
    chart = workbook.add_chart({'type': 'column'})
    chart.add_series({'values': [sheet_name, 1, 0, len(df_ordered_returns), 0]})
    chart.set_x_axis({'major_tick_mark': 'none',
                      'minor_tick_mark': 'none',
                      'label_position': 'none'})
    chart.set_y_axis({'name': 'Return',
                      'name_font': {'size': 14, 'bold': True},
                      'major_gridlines': {'visible': False},
                      'major_tick_mark': 'none',
                      'minor_tick_mark': 'none'})
    chart.set_title({'name': 'Daily returns from lowest to highest'})
    chart.set_legend({'position': 'none'})
    chartsheet = workbook.add_chartsheet('Chart Ordered returns')
    chartsheet.set_chart(chart)
    
# Plot the probability density function of daily returns by drawing an area chart using the chosen
# number of bins.
    sheet_name = out_sheets[5]
    worksheet = xl_writer.sheets[sheet_name]
    
    chart = workbook.add_chart({'type': 'area'})
    chart.add_series({'categories': [sheet_name, 1, 0, len(df_binned_returns), 0],
                      'values': [sheet_name, 1, 1, len(df_binned_returns), 1]})
    chart.set_x_axis({'name': 'Return (bps)',
                      'name_font': {'size': 14, 'bold': True},
                      'major_gridlines': {'visible': True, 'line': {'width': 0.25, 'dash_type': 'dash'}},
                      'major_tick_mark': 'none',
                      'minor_tick_mark': 'none'})
    chart.set_y_axis({'name_font': {'size': 14, 'bold': True},
                      'major_gridlines': {'visible': True, 'line': {'width': 0.25, 'dash_type': 'dash'}},
                      'major_tick_mark': 'none',
                      'minor_tick_mark': 'none',
                      'label_position': 'none'})
    chart.set_title({'name': 'PDF of daily returns'})
    chart.set_legend({'position': 'none'})
    chartsheet = workbook.add_chartsheet('Chart PDF returns')
    chartsheet.set_chart(chart)
    
# Plot the cumulative distribution function of daily returns by drawing a line chart using the chosen
# number of bins.
    sheet_name = out_sheets[5]
    worksheet = xl_writer.sheets[sheet_name]
    
    chart = workbook.add_chart({'type': 'line'})
    chart.add_series({'categories': [sheet_name, 1, 0, len(df_binned_returns), 0],
                      'values': [sheet_name, 1, 2, len(df_binned_returns), 2]})
    chart.set_x_axis({'name': 'Return (bps)',
                      'name_font': {'size': 14, 'bold': True},
                      'major_gridlines': {'visible': True, 'line': {'width': 0.25, 'dash_type': 'dash'}},
                      'major_tick_mark': 'none',
                      'minor_tick_mark': 'none'})
    chart.set_y_axis({'name_font': {'size': 14, 'bold': True},
                      'major_gridlines': {'visible': True, 'line': {'width': 0.25, 'dash_type': 'dash'}},
                      'major_tick_mark': 'none',
                      'minor_tick_mark': 'none',
                      'min': 0,
                      'max': 1})
    chart.set_title({'name': 'CDF of daily returns'})
    chart.set_legend({'position': 'none'})
    chartsheet = workbook.add_chartsheet('Chart CDF returns')
    chartsheet.set_chart(chart)
    
# Plot the Sharpe ratios since the start.
    sheet_name = out_sheets[6]
    worksheet = xl_writer.sheets[sheet_name]
    line_width = np.round((1500 / len(df_sharpe) * 2.25) / 0.25, 0) * 0.25
    
    if line_width > 3:
        line_width = 3
    elif line_width < 0.25:
        line_width = 0.25
    
    if sharpe_exclude < 0:
        sharpe_exclude = 0
    
    if sharpe_exclude < len(df_sharpe):
        chart = workbook.add_chart({'type': 'line'})
        chart.add_series({'categories': [sheet_name, sharpe_exclude + 1, 0, len(df_sharpe), 0],
                          'values': [sheet_name, sharpe_exclude + 1, 3, len(df_sharpe), 3],
                          'line': {'width': line_width}})
        chart.set_x_axis({'name': 'Date',
                          'name_font': {'size': 14, 'bold': True},
                          'major_gridlines': {'visible': True, 'line': {'width': 0.25, 'dash_type': 'dash'}},
                          'major_tick_mark': 'none',
                          'minor_tick_mark': 'none'})
        chart.set_y_axis({'name': 'Sharpe ratio',
                          'name_font': {'size': 14, 'bold': True},
                          'major_gridlines': {'visible': True, 'line': {'width': 0.25, 'dash_type': 'dash'}},
                          'major_tick_mark': 'none',
                          'minor_tick_mark': 'none'})
        chart.set_title({'name': 'Sharpe ratios since start'})
        chart.set_legend({'position': 'none'})
        chartsheet = workbook.add_chartsheet('Chart Sharpe ratios')
        chartsheet.set_chart(chart)
    
    xl_writer.save()

